/**********************************************************************
 *	RC-Meter : Sample Main
 **********************************************************************
 *	ڑ:
      PD7(AIN1)   ===> --R10K--+
                               |
      PD6(AIN0)   ===> --R1K---+
                               |
      PD1         ===> --R220--+
                               |                          |
      PC0(ADC0)===> --+--R10K--+-------*------<PROBE>---+ |
                      |                                  R?K
                     === 0.047uF                          |
                      |           +----*------<PROBE>---+ |
                     GND         GND                      |


      PD0(RxD)    ===> N.C.

 **********************************************************************
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "config.h"
#include "task.h"
#include "taskdef.h"
#include "timer.h"


#define	R220_OFFSET	22			// AVRPULLUP R  220̒R덷̘a.
#define	R1K_OFFSET	22			// AVRPULLUP R   1K̒R덷̘a.
#define	R10K_OFFSET	150			// AVRPULLUP R  10K̒R덷̘a.

#define	PULLUP_OFFSET	4		// AVRPULLUP d VccƂ̌덷digit (5V/1024)
#define	PROBE_OFFSET	1		// v[u̓R.

/**********************************************************************
 *	define
 **********************************************************************
 */
void quit(void);				// bootloaderɖ߂.
void led_task(void);			// LED_hCo[
void led_set_number(int data);	// LED\f[^̃Zbg.
void led_set_pattern(uchar *data);// LED\p^[̃Zbg.
void led_set_comma(uchar digit);// ʒu2ڂɍ킹.
void led_set_zero_supress(void);// ʌ̃[.
void wait_usec(uchar n);
/**********************************************************************
 *	work
 **********************************************************************
 */
#define		   MAX_SAMPLE		16
#define		   MAX_SAMPLE_SFT	4
uint  ain_data[MAX_SAMPLE];
uchar ain_cnt = 0;
uint  ain_val;		// ϒl.

ulong Rval;			// Rl. Pʂ̓.
uint  r_val;		// Rl. (4ȓ)
uchar r_exp;		// Rl _ʒu.
uchar r_mode;		// R:݂PULL-UP[h(0=220,1=1K,2=10K)
uchar v_mode;		// 0= / 1=d / 255=N.
uchar v_cnt0;

enum {
	MODE_R220=0,
	MODE_R1K =1,
	MODE_R10K=2,
	MODE_VOLT=3,
};

#define	PIN_R220	1
#define	PIN_R1K		6
#define	PIN_R10K	7

/**********************************************************************
 *	A/Dϊ̎s.
 **********************************************************************
 */
static	int adc_exec(void)
{
	uchar c;
	uint  d;
	ADCSRA |= 0b0100_0000;	// ADSC (start)

	do {
		c = ADCSRA;
	} while(c & 0b0100_0000);
	//	ϊI
	d  = ADCL;
	d |=(ADCH<<8);
	return d;
}

/**********************************************************************
 *	A/Dϊ task.
 **********************************************************************
 */
static	void adc_task(void)
{
	uint d = adc_exec();
	ain_data[ain_cnt] = d;
	ain_cnt++;
	if(	ain_cnt >= MAX_SAMPLE) {
		ain_cnt = 0;
	}
}

/**********************************************************************
 *	PORTD ̎wbit  Hi xo͂ɂ.
 **********************************************************************
 */
static	void set_HiD(uchar bit)
{
	uchar d = 1<<bit;
	PORTD |=  d;			// Hi
	DDRD  |=  d;			// PDx: H
}
/**********************************************************************
 *	PORTD ̎wbit  I[v ɂ.
 **********************************************************************
 */
static	void set_HiZ(uchar bit)
{
	uchar d = ~(1<<bit);
	PORTD &= d;				//      Lo
	DDRD  &= d;				// PDx: Hi-Z
}

/**********************************************************************
 *	PULL UP [h؂ւ.
 **********************************************************************
 */
static	void set_pullup_mode(uchar mode)
{
	uchar pintbl[4]={PIN_R220,PIN_R1K,PIN_R10K,PIN_R10K};
	uchar i;

	r_mode = mode;

	for(i=0;i<3;i++) {
		if(i==mode) {
			set_HiD(pintbl[i]);
		}else{
			set_HiZ(pintbl[i]);
		}
	}
}

int get_pullup_R(void)
{
	uint regtbl[3]={
		  220+R220_OFFSET,
		 1000+R1K_OFFSET,
		10000+R10K_OFFSET,
	};
	return regtbl[r_mode];
}
/**********************************************************************
 *	PULL UP [hŐ؂ւ.
 **********************************************************************
 */
#define	R10000 5000
#define	R1000  500
static void change_mode(void)
{
	uchar mode;
	if(r_exp == 255)   {mode = MODE_R10K; }else 
	if(Rval >= R10000) {mode = MODE_R10K; }else 
	if(Rval >= R1000 ) {mode = MODE_R1K;  }else 
    { mode = MODE_R220; }

	if(r_mode != mode) {
		set_pullup_mode(mode);
	}
}
/**********************************************************************
 *	`/c ͂̕ς.
 **********************************************************************
 */
static void measure(void)
{
	uint d=0;
	uchar i;
	for(i=0;i<MAX_SAMPLE;i++ ) {
		d += ain_data[i];
	}
	d >>= MAX_SAMPLE_SFT;	// 16Ŋ.
	ain_val = d;
}
/**********************************************************************
 *	`/c ͂̕ς.
 **********************************************************************
 */
static void measure_volt(void)
{
	Rval = ain_val;
	Rval = (500 * Rval ) / 1024;	// 5.000 V
	r_val = Rval;
	r_exp = 2;
}
/**********************************************************************
 *	RlvZ.
 **********************************************************************

                     Vin
	Rval = Rref * ---------
                  Vcc - Vin
 */
static void calculate(void)
{
	ulong Vin;
	ulong Vcc;
	ulong Rref;

	Rref = get_pullup_R();
	if( ain_val >=(1023 - PULLUP_OFFSET )) {
		r_val = 9999;
		r_exp = 255;
		return ;
	}
	Vin =    ain_val;
	Vcc = (1023 - PULLUP_OFFSET ) - ain_val ;

	Rval = ( Vin * Rref ) / Vcc;
	if(	Rval>=PROBE_OFFSET) {
		Rval-=PROBE_OFFSET;
	}

	r_exp = 0;				//܂A_ 0 ڂƂ.
	if(Rval < 1000) {		// 1k.
		r_exp = 3;
		r_val = Rval;
		return;
	}
	
	if(Rval < 10000) {		// 10k.
		r_exp = 2;
		r_val = (Rval/10);
		return;
	}
	
	if(Rval < 100000) {		// 100k.
		r_exp = 1;
		r_val = (Rval/100);
		return;
	}
	
	if(Rval < 1000000) {		// 1000k.
		r_exp = 0;
		r_val = (Rval/1000);
		return;
	}
	
		r_val = 9999;
		r_exp = 255;
		return ;
}
/**********************************************************************
 *	result		ʕ\.
 **********************************************************************
 */
static void disp_result()
{
	static	uchar minus_data[4]={
		0b01000000,	// -
		0b01000000,	// -
		0b01000000,	// -
		0b01000000,	// -
	};

	if(r_exp==255) {
//		led_set_number(9999);		//I[v\9999
		led_set_pattern(minus_data);
		return;
	}

	led_set_number(r_val);	// LEDɐݒ肷.
	led_set_comma(r_exp);	// ʒu2ڂɍ킹.
	led_set_zero_supress();	// ʌ̃[.
}
/**********************************************************************
 *	task
 **********************************************************************
 */
static void volt_task(void)
{
	measure();
	measure_volt();
	disp_result();	//ʂLEDɃZbg.
	sleep_tsk(48);	//128 mSx.
}
/**********************************************************************
 *	d[hֈڍs?
 **********************************************************************
 */
static void check_volt_mode(void)
{
	if(	v_mode == 255) {	//[h?
		v_cnt0++;
		if(v_cnt0==3) {		//NRڂ̌v.
			if(r_exp != 3) {	// 1kłȂ.
				v_mode = 0;		// ohm_mode ɌŒ;
			}else{				// 1kȏ.
				v_mode = 1;		// volt_modeɌŒ;
				set_pullup_mode(MODE_VOLT);		// ohmpullupO.
			}
		}
	}
}
/**********************************************************************
 *	task
 **********************************************************************
 */
static void ohm_task(void)
{
	measure();
	calculate();
	disp_result();	//ʂLEDɃZbg.

	change_mode();
	check_volt_mode();
	sleep_tsk(48);	//128 mSx.
}
/**********************************************************************
 *	task
 **********************************************************************
 */
static void main_task(void)
{
	if(v_mode==1) {
		volt_task();
	}else{
		ohm_task();
	}
}
/**********************************************************************
 *	task
 **********************************************************************
 */
static void task_regist(void)
{
	reg_tsk(ID_adc, adc_task);
	reg_tsk(ID_main,main_task);
	reg_tsk(ID_led, led_task);
}

/**********************************************************************
 *	
 **********************************************************************
 */
static void port_init(void)
{
	//|[gWX^̏.
	PORTC=0;
	DDRC =0b1111_1110;		// PC0 == AIN0

	PORTD=0b0011_0000;		// PD7,PD6,PD1 == R10K,R1K,R220
	DDRD =0b0011_0000;

	PORTB=0;
	DDRB =0b0011_1111;

	// A/Dϊ MUXWX^ݒ.
	ADMUX = 0b0000_0000;
	// A/Dϊ dOn, A/D Clock = 1/64 ck
	ADCSRA = 0b1000_0110;
	// AINɑ΂fW^͋֎~.
	DIDR0  = 0b0000_0001;


#if	0	// rgp鏉.
	//A/DϊpWX^ݒ.
	ADCSRA = 0b1000_0000;
	//AiOr평.
	ADCSRB = 0b0000_0000;
	//d1.1VƔr / Timer1ߊlOn. / GbWł̊荞.
	ACSR   = 0b0100_0110;
	//AIN1ɑ΂fW^͋֎~.
	DIDR1  = 0b0000_0010;
#endif
}
/**********************************************************************
 *	C
 **********************************************************************
 */
int	main(void)
{
	port_init();
						// ȉA20MHz̎:
	timer_init(41666-1);// 2083S(480.0  Hz)Ŋ荞.
//	timer_init(256-1);	// 12.8S(78.125kHz)Ŋ荞.
	sei();

	task_regist();
	set_pullup_mode(MODE_R1K);
	v_mode = 255;		// [h.
	v_cnt0 = 0;
	do {
		timer_wait();
		dispatch_tsk();
	}while(1);

	return 0;
}

/**********************************************************************
 *	
 **********************************************************************
 */
